/**
 * <copyright>
 *
 * Service Data Objects
 * Version 3.0
 * Licensed Materials
 *
 * (c) Copyright BEA Systems, Inc., International Business Machines Corporation,
 * Oracle Corporation, Primeton Technologies Ltd., Rogue Wave Software, SAP AG.,
 * Software AG., Sun Microsystems, Sybase Inc., Xcalia, Zend Technologies,
 * 2005-2008. All rights reserved.
 *
 * </copyright>
 *
 */
package commonj.sdo.impl;

import java.lang.ref.WeakReference;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import commonj.sdo.helper.HelperContext;
import commonj.sdo.helper.HelperContextFactory;

/**
 * This class allows SDO to be configured for a specific environment.  Possible environments
 * include:
 * - Java SE
 * - Application Servers
 * - OSGi environments
 */
public abstract class Environment {

    private static Environment environment;

    public static Environment getEnvironment() {
        if (null == environment) {
            environment = ServiceResolver.getEnvironment();
        }
        return environment;
    } 
    
    public static void setEnvironment(Environment env) {
        environment = env;
    }

    /**
     * Return the default HelperContextFactory.
     */
    public abstract HelperContextFactory getHelperContextFactory();
    
    /**
     * Set the default HelperContextFactory.
     */
    public abstract void setHelperContextFactory(HelperContextFactory factory);

    /**
     * Return a non-default HelperContextFactory with the specified implementation class.
     */
    public abstract HelperContextFactory getHelperContextFactory(String implName);

    /**
     * Add a named HelperContextFactory, which can be accessed using it's implementation class name.
     */
    public abstract void addHelperContextFactory(HelperContextFactory factory);

    /**
     * Return the default HelperContext.  The HelperContext returned must have
     * null as the identifier.
     */
    public abstract HelperContext getDefaultHelperContext();

    /**
     * Return the HelperContext with specified identifier, if a HelperContext
     * has been registered with that identifier.  Calling this method with
     * a null identifier is equivalent to SDO.getDefaultHelperContext().
     */
    public abstract HelperContext getHelperContext(String identifier);

    /**
     * Register a HelperContext.
     * @return true if the HelperContext was successfully added (would not be added if a different HelperContext was registered with that identifier.
     */
    public abstract boolean addHelperContext(HelperContext helperContext);
    
    /**
     * Default implementation of Environment.
     */
    public static class DefaultImpl extends Environment
    {
        protected ConcurrentMap<String, WeakReference<HelperContextFactory>> helperContextFactories;
        protected ConcurrentMap<String, WeakReference<HelperContext>> helperContexts;
        protected HelperContextFactory defaultContextFactory = null;
        protected HelperContext defaultContext = null;

        public DefaultImpl()
        {
            helperContextFactories = new ConcurrentHashMap<String, WeakReference<HelperContextFactory>>();
            helperContexts = new ConcurrentHashMap<String, WeakReference<HelperContext>>();
        }
        
        public HelperContextFactory getHelperContextFactory() {
            if (defaultContextFactory == null) {
                defaultContextFactory = ServiceResolver.getHelperContextFactory(this);
                addHelperContextFactory(defaultContextFactory);
            }
            return defaultContextFactory;
        }
        
        public void setHelperContextFactory(HelperContextFactory factory) {
            defaultContextFactory = factory;
            addHelperContextFactory(defaultContextFactory);
        }

        public HelperContextFactory getHelperContextFactory(String implName) {
            if (helperContextFactories.containsKey(implName)) {
                return helperContextFactories.get(implName).get();
            }
            HelperContextFactory factory = ServiceResolver.getHelperContextFactory(this, implName);
            helperContextFactories.put(implName, new WeakReference<HelperContextFactory>(factory));
            return factory;
        }
        
        public void addHelperContextFactory(HelperContextFactory factory) {
            String implName = factory.getClass().getName();
            helperContextFactories.put(implName, new WeakReference<HelperContextFactory>(factory));         
        }

        public HelperContext getDefaultHelperContext() {
            if (defaultContext == null) {
                defaultContext = getHelperContextFactory().createHelperContext(null);
            }
            return defaultContext;
        }

        public HelperContext getHelperContext(String identifier) {
            return helperContexts.get(identifier).get();
        }

        public boolean addHelperContext(HelperContext helperContext) {
            return helperContexts.putIfAbsent(helperContext.getIdentifier(), new WeakReference<HelperContext>(helperContext)) == null;
        }
    }
    
    /**
     * Base implementation of HelperContextFactory.
     */
    public static abstract class BaseHelperContextFactory implements HelperContextFactory
    {
        private static long helperContextNumber = 0;       
        protected final Environment environment;
        
        public BaseHelperContextFactory(Environment environment)
        {
            this.environment = environment;
        }
        
        public HelperContext createHelperContext(Map<String, Object> properties) throws IllegalArgumentException
        {
            return createHelperContext(properties, "HelperContextID" + ++helperContextNumber);
        }
        
        public HelperContext createHelperContext(Map<String, Object> properties, String identifier) throws IllegalArgumentException
        {
            HelperContext hc = basicCreate(properties, identifier);
            if (null != identifier && !"".equals(identifier)) {
                if (!environment.addHelperContext(hc)) throw new IllegalArgumentException();
            }
            return hc;
        }
        
        protected abstract HelperContext basicCreate(Map<String, Object> properties, String identifier) throws IllegalArgumentException;
    }
    
}
